import json
from typing import cast
from unittest import IsolatedAsyncioTestCase as TestCase
from unittest.mock import Mock, call, patch
from urllib.parse import urlencode

from fastapi import Request
from oic.oic import AuthorizationRequest, Client as OIDCClient
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from regal import Address
from verifier import ContactData
from verifier.exceptions import (
    IncorrectStatus,
    InvalidData,
    MethodNotApplicable,
    VerificationDoesNotExist,
)

from inquisitor.constants import LogEntryType, LoggerResult, LogReferenceType, VerificationMethod
from inquisitor.main import app
from inquisitor.models import Error
from inquisitor.oidc import OIDCAuthResult, get_oidc_client
from inquisitor.settings import OIDCSettings, Settings

from .utils import (
    CONTACT_ID,
    VERIFICATION_ID,
    ApiTestCase,
    make_verification,
    make_verification_result,
)


class OIDCClientMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.oidc_client = Mock(spec=OIDCClient)
        self.oidc_client.client_id = "JMC"
        self.oidc_client.client_secret = "Gazpacho!"
        self.oidc_client.redirect_uris = ["https://localhost/oidc/token/"]
        self.oidc_client.provider_config.return_value = None
        app.dependency_overrides[get_oidc_client] = lambda: self.oidc_client

    def tearDown(self):
        cast(TestCase, super()).tearDown()
        app.dependency_overrides = {}


class GetOidcClientTest(TestCase):
    async def test_get_oidc_client(self):
        request = Request(scope={"type": "http", "headers": {}, "router": app.router})
        with patch("inquisitor.oidc.OIDCClient") as mock:
            mock.provider_config.return_value = None
            client = get_oidc_client(
                request,
                Settings(
                    oidc=OIDCSettings(
                        issuer="https://example.com/oidc/",
                        client_id="JMC",
                        client_secret="Gazpacho!",
                    )
                ),
            )

        self.assertEqual(client.client_id, "JMC")
        self.assertEqual(client.client_secret, "Gazpacho!")
        self.assertEqual(
            mock.mock_calls,
            [
                call(client_authn_method=CLIENT_AUTHN_METHOD),
                call().provider_config("https://example.com/oidc/"),
            ],
        )


class OIDCAuthTest(OIDCClientMixin, ApiTestCase):
    async def _test_auth(
        self,
        *,
        redirect_to: str,
        log_result: LoggerResult = LoggerResult.SUCCESS,
        output_properties: dict | None = None,
    ) -> None:
        self.oidc_client.authorization_endpoint = "https://example.org/oidc/auth/"
        auth_req = Mock(spec=AuthorizationRequest)
        auth_req.request.return_value = "https://example.org/oidc/auth/?some=query"
        self.oidc_client.construct_AuthorizationRequest.return_value = auth_req

        response = self.client.get(
            "/oidc/auth/?" + urlencode({"redirect": "/custom/redirect/"}), follow_redirects=False
        )

        self.assertTrue(response.is_redirect)
        self.assertEqual(response.headers["Location"], redirect_to)
        self.assertEqual(
            self.verification_client.mock_calls,
            [call.use_method(VERIFICATION_ID, method=VerificationMethod.MOJEID)],
        )
        self.assertLog(
            LogEntryType.USE_METHOD,
            log_result,
            input_properties={"method": VerificationMethod.MOJEID},
            output_properties=output_properties,
            input_references={
                LogReferenceType.VERIFICATION: VERIFICATION_ID,
                LogReferenceType.CONTACT: CONTACT_ID,
            },
        )

    async def test_auth(self):
        await self._test_auth(
            redirect_to="https://example.org/oidc/auth/?some=query",
        )

    async def _test_use_method_error(self, detail: str) -> None:
        await self._test_auth(
            redirect_to="/custom/redirect/",
            log_result=LoggerResult.FAIL,
            output_properties={"error": detail},
        )

    async def test_use_method_verification_does_not_exist(self):
        self.verification_client.use_method.side_effect = VerificationDoesNotExist
        await self._test_use_method_error("Verification not found")

    async def test_use_method_not_applicable(self):
        self.verification_client.use_method.side_effect = MethodNotApplicable
        await self._test_use_method_error("Method not applicable")

    async def test_use_method_incorrect_status(self):
        self.verification_client.use_method.side_effect = IncorrectStatus
        await self._test_use_method_error("Incorrect verification status")


class OIDCTokenTest(OIDCClientMixin, ApiTestCase):
    async def test_token(self):
        oidc_response = {
            "state": self.session.oidc.state,
            "code": "Gazpacho!",
        }
        userinfo = {
            "name": "Arnold Rimmer",
            "mojeid_nia": True,
            "mojeid_valid": False,
        }
        contact_data = ContactData(name="Arnold Rimmer", organization="")
        self.oidc_client.parse_response.return_value = oidc_response
        self.oidc_client.do_access_token_request.return_value = None
        self.oidc_client.do_user_info_request.return_value = userinfo
        self.verification_client.verify_contact_data.return_value = (
            make_verification(),
            make_verification_result(data=contact_data),
        )

        response = self.client.get("/oidc/token/", follow_redirects=False)

        self.assertTrue(response.is_redirect)
        self.assertRegex(response.headers["Location"], "/")
        self.assertEqual(
            self.verification_client.mock_calls,
            [
                call.verify_contact_data(
                    VERIFICATION_ID, method=VerificationMethod.MOJEID, data=contact_data
                )
            ],
        )
        self.assertLog(
            LogEntryType.VERIFY_CONTACT_DATA,
            LoggerResult.SUCCESS,
            input_properties={"method": VerificationMethod.MOJEID},
            input_references={
                LogReferenceType.VERIFICATION: VERIFICATION_ID,
                LogReferenceType.CONTACT: CONTACT_ID,
            },
            input_content=json.dumps(oidc_response, indent=4, ensure_ascii=False),
            output_content=json.dumps(userinfo, indent=4, ensure_ascii=False),
        )

    async def test_token_not_trustworthy(self):
        oidc_response = {
            "state": self.session.oidc.state,
            "code": "Gazpacho!",
        }
        userinfo = {
            "mojeid_organization": "Jupiter Mining Corporation",
            "mojeid_address_def_street": "Deck 5",
            "mojeid_address_def_street2": "Cabin 101",
            "mojeid_address_def_city": "Red Dwarf",
            "mojeid_address_def_postal_code": "JMC",
            "mojeid_address_def_country": "CZ",
            "mojeid_nia": False,
            "mojeid_valid": False,
        }
        contact_data = ContactData(
            name="",
            organization="Jupiter Mining Corporation",
            address=Address(
                street=["Deck 5", "Cabin 101"],
                city="Red Dwarf",
                postal_code="JMC",
                country_code="CZ",
            ),
        )
        self.oidc_client.parse_response.return_value = oidc_response
        self.oidc_client.do_access_token_request.return_value = None
        self.oidc_client.do_user_info_request.return_value = userinfo
        self.verification_client.verify_contact_data.return_value = (
            make_verification(),
            make_verification_result(data=contact_data),
        )

        response = self.client.get("/oidc/token/", follow_redirects=False)

        self.assertTrue(response.is_redirect)
        self.assertRegex(response.headers["Location"], "/")
        self.assertEqual(self.verification_client.mock_calls, [])
        self.assertLog(
            LogEntryType.VERIFY_CONTACT_DATA,
            LoggerResult.FAIL,
            input_properties={"method": VerificationMethod.MOJEID},
            input_references={
                LogReferenceType.VERIFICATION: VERIFICATION_ID,
                LogReferenceType.CONTACT: CONTACT_ID,
            },
            input_content=json.dumps(oidc_response, indent=4, ensure_ascii=False),
            output_content=json.dumps(userinfo, indent=4, ensure_ascii=False),
        )

    async def test_token_wrong_state(self):
        oidc_response = {
            "state": "wrong state",
            "code": "Gazpacho!",
        }
        self.oidc_client.parse_response.return_value = oidc_response

        response = self.client.get("/oidc/token/", follow_redirects=False)

        self.assertEqual(response.status_code, 401)
        self.assertLog(
            LogEntryType.VERIFY_CONTACT_DATA,
            LoggerResult.FAIL,
            input_properties={"method": VerificationMethod.MOJEID},
            input_references={
                LogReferenceType.VERIFICATION: VERIFICATION_ID,
                LogReferenceType.CONTACT: CONTACT_ID,
            },
            input_content=json.dumps(oidc_response, indent=4, ensure_ascii=False),
        )

    async def _test_token_error(self, detail: str) -> None:
        oidc_response = {
            "state": self.session.oidc.state,
            "code": "Gazpacho!",
        }
        userinfo = {
            "name": "Arnold Rimmer",
            "mojeid_nia": True,
            "mojeid_valid": False,
        }
        contact_data = ContactData(name="Arnold Rimmer", organization="")
        self.oidc_client.parse_response.return_value = oidc_response
        self.oidc_client.do_access_token_request.return_value = None
        self.oidc_client.do_user_info_request.return_value = userinfo
        self.verification_client.verify_contact_data.return_value = (
            make_verification(),
            make_verification_result(data=contact_data),
        )

        response = self.client.get("/oidc/token/", follow_redirects=False)

        self.assertTrue(response.is_redirect)
        self.assertRegex(response.headers["Location"], "/")
        self.assertEqual(
            self.verification_client.mock_calls,
            [
                call.verify_contact_data(
                    VERIFICATION_ID, method=VerificationMethod.MOJEID, data=contact_data
                )
            ],
        )
        self.assertLog(
            LogEntryType.VERIFY_CONTACT_DATA,
            LoggerResult.FAIL,
            input_properties={"method": VerificationMethod.MOJEID},
            input_references={
                LogReferenceType.VERIFICATION: VERIFICATION_ID,
                LogReferenceType.CONTACT: CONTACT_ID,
            },
            output_properties={"error": detail},
            input_content=json.dumps(oidc_response, indent=4, ensure_ascii=False),
            output_content=json.dumps(userinfo, indent=4, ensure_ascii=False),
        )

    async def test_oidc_error(self):
        oidc_response = {
            "state": self.session.oidc.state,
            "error": "access_denied",
        }
        self.oidc_client.parse_response.return_value = oidc_response

        response = self.client.get("/oidc/token/", follow_redirects=False)

        self.assertTrue(response.is_redirect)
        self.assertRegex(response.headers["Location"], "/")
        self.assertEqual(self.verification_client.mock_calls, [])
        self.assertLog(
            LogEntryType.VERIFY_CONTACT_DATA,
            LoggerResult.FAIL,
            input_properties={"method": VerificationMethod.MOJEID},
            input_references={
                LogReferenceType.VERIFICATION: VERIFICATION_ID,
                LogReferenceType.CONTACT: CONTACT_ID,
            },
            input_content=json.dumps(oidc_response, indent=4, ensure_ascii=False),
            output_properties={"error": "access_denied"},
        )

    async def test_token_verification_does_not_exist(self):
        self.verification_client.verify_contact_data.side_effect = VerificationDoesNotExist
        await self._test_token_error("Verification not found")

    async def test_token_not_applicable(self):
        self.verification_client.verify_contact_data.side_effect = MethodNotApplicable
        await self._test_token_error("Method not applicable")

    async def test_token_incorrect_status(self):
        self.verification_client.verify_contact_data.side_effect = IncorrectStatus
        await self._test_token_error("Incorrect verification status")

    async def test_token_invalid_data(self):
        self.verification_client.verify_contact_data.side_effect = InvalidData(
            fields=["birthdate"]
        )
        await self._test_token_error("Invalid data")


class GetLastOIDCResultTest(ApiTestCase):
    def test_get_last_oidc_result(self):
        contact_data = ContactData(name="Arnold Rimmer", organization="")
        self.session.oidc.contact_data = contact_data
        self.session.oidc.mojeid_nia = True
        self.session.oidc.mojeid_validation = False
        self.session.oidc.verification_result = make_verification_result()

        response = self.client.get("/oidc/last/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            OIDCAuthResult(**response.json()),
            OIDCAuthResult(
                contact_data=contact_data,
                mojeid_nia=True,
                mojeid_validation=False,
                is_legal_person=False,
                verification_result=make_verification_result(),
            ),
        )

    def test_get_last_oidc_result_not_available(self):
        self.session.oidc.contact_data = None

        response = self.client.get("/oidc/last/")

        self.assertEqual(response.status_code, 400)

    def test_get_last_oidc_result_error(self):
        self.session.oidc.error = Error(detail="Incorrect status")

        response = self.client.get("/oidc/last/")

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Incorrect status"})
