import tempfile
from unittest import TestCase
from unittest.mock import call, mock_open, patch, sentinel

import yaml
from testfixtures import LogCapture

from inquisitor.constants import LOGGER_SERVICE, LoggerResult
from inquisitor.settings import (
    Settings,
    get_cache,
    get_contact_client,
    get_domain_client,
    get_logger,
    get_logger_client,
    get_registrar_client,
    get_verification_client,
)

from .utils import set_environ

REQUIRED_SETTINGS = {
    "secret_key": "Gazpacho!",
    "registry_netloc": "registry.example.org",
    "logger_netloc": "logger.example.org",
    "verification_netloc": "verification.example.org",
    "redis_cache_cls": "redis.cache.Class",
    "oidc": {
        "issuer": "https://example.org/oidc/",
        "client_id": "JMC",
        "client_secret": "Gazpacho!",
    },
}
REQUIRED_SETTINGS_ENV = {
    "INQUISITOR_SECRET_KEY": "Gazpacho!",
    "INQUISITOR_REGISTRY_NETLOC": "registry.example.org",
    "INQUISITOR_LOGGER_NETLOC": "logger.example.org",
    "INQUISITOR_VERIFICATION_NETLOC": "verification.example.org",
    "INQUISITOR_REDIS_CACHE_CLS": "redis.cache.Class",
    "INQUISITOR_OIDC__ISSUER": "https://example.org/oidc/",
    "INQUISITOR_OIDC__CLIENT_ID": "JMC",
    "INQUISITOR_OIDC__CLIENT_SECRET": "Gazpacho!",
}


class SettingsTest(TestCase):
    home_call = call("/home/holly/.fred/inquisitor.conf")
    system_call = call("/etc/fred/inquisitor.conf")

    def setUp(self):
        self.log_handler = LogCapture("inquisitor", propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()

    def test_config_file_environ(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, "w") as config_io:
            yaml.dump(REQUIRED_SETTINGS, config_io)

        with set_environ(INQUISITOR_CONFIG=config_file.name):
            settings = Settings()

        self.assertEqual(settings.verification_netloc, "verification.example.org")

    def test_config_file_environ_unknown(self):
        with set_environ(INQUISITOR_CONFIG="/does/not/exist"):
            with self.assertRaisesRegex(FileNotFoundError, "/does/not/exist"):
                Settings()

    def test_config_file_argv(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, "w") as config_io:
            yaml.dump(REQUIRED_SETTINGS, config_io)

        with set_environ():
            settings = Settings(config_file=config_file.name)

        self.assertEqual(settings.verification_netloc, "verification.example.org")

    def test_config_file_argv_unknown(self):
        with set_environ():
            with self.assertRaisesRegex(FileNotFoundError, "/does/not/exist"):
                Settings(config_file="/does/not/exist")

    def test_config_file_home(self):
        data = yaml.dump(REQUIRED_SETTINGS)

        with set_environ(HOME="/home/holly"):
            with patch("os.path.isfile", return_value=True):
                with patch("builtins.open", mock_open(read_data=data)) as open_mock:
                    settings = Settings()

        self.assertEqual(settings.verification_netloc, "verification.example.org")
        self.assertIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)

    def test_config_file_system(self):
        data = yaml.dump(REQUIRED_SETTINGS)

        with set_environ(HOME="/home/holly"):
            with patch("os.path.isfile", side_effect=[False, True]):
                with patch("builtins.open", mock_open(read_data=data)) as open_mock:
                    settings = Settings()

        self.assertEqual(settings.verification_netloc, "verification.example.org")
        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertIn(self.system_call, open_mock.mock_calls)

    def test_config_file_no_settings(self):
        with set_environ(HOME="/home/holly", **REQUIRED_SETTINGS_ENV):
            with patch("os.path.isfile", return_value=False):
                with patch("builtins.open", side_effect=FileNotFoundError) as open_mock:
                    Settings()

        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)
        # Check no logs were made.
        self.log_handler.check()

    def test_setup_logging(self):
        config = {"logging_key": "logging_value"}
        with patch("inquisitor.settings.dictConfig") as config_mock:
            Settings(logging=config)

        self.assertEqual(config_mock.mock_calls, [call(config)])

    def test_setup_no_sentry(self):
        with patch("inquisitor.settings.sentry_sdk") as sentry_mock:
            Settings()

        self.assertEqual(sentry_mock.mock_calls, [])

    def test_setup_sentry(self):
        dsn = "SMEGHEAD!"
        with patch("inquisitor.settings.sentry_sdk") as sentry_mock:
            Settings(sentry={"dsn": dsn})

        self.assertEqual(
            sentry_mock.mock_calls,
            [call.init(dsn=dsn, environment=None, ca_certs=None)],
        )

    def test_setup_sentry_missing(self):
        dsn = "SMEGHEAD!"
        with patch("inquisitor.settings.sentry_sdk", new=None):
            Settings(sentry={"dsn": dsn})

        self.log_handler.check(
            (
                "inquisitor.settings",
                "WARNING",
                "Sentry set up, but sentry-sdk not available.",
            )
        )

    def test_setup_sentry_environment(self):
        dsn = "SMEGHEAD!"
        environment = "JMC"
        with patch("inquisitor.settings.sentry_sdk") as sentry_mock:
            Settings(sentry={"dsn": dsn, "environment": environment})

        self.assertEqual(
            sentry_mock.mock_calls,
            [call.init(dsn=dsn, environment=environment, ca_certs=None)],
        )

    def test_setup_sentry_ca_certs(self):
        dsn = "SMEGHEAD!"
        ca_certs = "Silver Swimming Certificate"
        with patch("inquisitor.settings.sentry_sdk") as sentry_mock:
            Settings(sentry={"dsn": dsn, "ca_certs": ca_certs})

        self.assertEqual(
            sentry_mock.mock_calls,
            [call.init(dsn=dsn, environment=None, ca_certs=ca_certs)],
        )


class GetContactClientDependencyTest(TestCase):
    def test_get_contact_client(self):
        settings = Settings(registry_netloc="example.org")
        with patch("inquisitor.settings.ContactClient") as client_mock:
            client_mock.return_value = sentinel.client
            self.assertEqual(get_contact_client(settings), sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call("example.org", credentials=None)])

    def test_get_contact_client_credentials(self):
        settings = Settings(registry_netloc="example.org", registry_ssl_cert="ssl-cert")
        with patch("inquisitor.settings.ContactClient") as client_mock:
            with patch("inquisitor.settings.make_credentials") as cred_mock:
                client_mock.return_value = sentinel.client
                cred_mock.return_value = sentinel.credentials
                self.assertEqual(get_contact_client(settings), sentinel.client)

        self.assertEqual(
            client_mock.mock_calls, [call("example.org", credentials=sentinel.credentials)]
        )
        self.assertEqual(cred_mock.mock_calls, [call("ssl-cert")])


class GetDomainClientDependencyTest(TestCase):
    def test_get_domain_client(self):
        settings = Settings(registry_netloc="example.org")
        with patch("inquisitor.settings.DomainClient") as client_mock:
            client_mock.return_value = sentinel.client
            self.assertEqual(get_domain_client(settings), sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call("example.org", credentials=None)])

    def test_get_domain_client_credentials(self):
        settings = Settings(registry_netloc="example.org", registry_ssl_cert="ssl-cert")
        with patch("inquisitor.settings.DomainClient") as client_mock:
            with patch("inquisitor.settings.make_credentials") as cred_mock:
                client_mock.return_value = sentinel.client
                cred_mock.return_value = sentinel.credentials
                self.assertEqual(get_domain_client(settings), sentinel.client)

        self.assertEqual(
            client_mock.mock_calls, [call("example.org", credentials=sentinel.credentials)]
        )
        self.assertEqual(cred_mock.mock_calls, [call("ssl-cert")])


class GetRegistrarClientDependencyTest(TestCase):
    def test_get_registrar_client(self):
        settings = Settings(registry_netloc="example.org")
        with patch("inquisitor.settings.RegistrarClient") as client_mock:
            client_mock.return_value = sentinel.client
            self.assertEqual(get_registrar_client(settings), sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call("example.org", credentials=None)])

    def test_get_registrar_client_credentials(self):
        settings = Settings(registry_netloc="example.org", registry_ssl_cert="ssl-cert")
        with patch("inquisitor.settings.RegistrarClient") as client_mock:
            with patch("inquisitor.settings.make_credentials") as cred_mock:
                client_mock.return_value = sentinel.client
                cred_mock.return_value = sentinel.credentials
                self.assertEqual(get_registrar_client(settings), sentinel.client)

        self.assertEqual(
            client_mock.mock_calls, [call("example.org", credentials=sentinel.credentials)]
        )
        self.assertEqual(cred_mock.mock_calls, [call("ssl-cert")])


class GetVerificationClientDependencyTest(TestCase):
    def test_get_verification_client(self):
        settings = Settings(verification_netloc="example.org")
        with patch("inquisitor.settings.VerificationClient") as client_mock:
            client_mock.return_value = sentinel.client
            self.assertEqual(get_verification_client(settings), sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call("example.org", credentials=None)])

    def test_get_verification_client_credentials(self):
        settings = Settings(verification_netloc="example.org", verification_ssl_cert="ssl-cert")
        with patch("inquisitor.settings.VerificationClient") as client_mock:
            with patch("inquisitor.settings.make_credentials") as cred_mock:
                client_mock.return_value = sentinel.client
                cred_mock.return_value = sentinel.credentials
                self.assertEqual(get_verification_client(settings), sentinel.client)

        self.assertEqual(
            client_mock.mock_calls, [call("example.org", credentials=sentinel.credentials)]
        )
        self.assertEqual(cred_mock.mock_calls, [call("ssl-cert")])


class GetLoggerClientDependencyTest(TestCase):
    def test_get_logger_client(self):
        settings = Settings(logger_netloc="example.org")
        with patch("inquisitor.settings.LoggerClient") as client_mock:
            client_mock.return_value = sentinel.client
            self.assertEqual(get_logger_client(settings), sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call("example.org", credentials=None)])

    def test_get_logger_client_credentials(self):
        settings = Settings(logger_netloc="example.org", logger_ssl_cert="ssl-cert")
        with patch("inquisitor.settings.LoggerClient") as client_mock:
            with patch("inquisitor.settings.make_credentials") as cred_mock:
                client_mock.return_value = sentinel.client
                cred_mock.return_value = sentinel.credentials
                self.assertEqual(get_logger_client(settings), sentinel.client)

        self.assertEqual(
            client_mock.mock_calls, [call("example.org", credentials=sentinel.credentials)]
        )
        self.assertEqual(cred_mock.mock_calls, [call("ssl-cert")])


class GetLoggerDependencyTest(TestCase):
    def test_get_logger(self):
        with patch("inquisitor.settings.Logger") as mock:
            mock.return_value = sentinel.logger
            self.assertEqual(get_logger(sentinel.client), sentinel.logger)

        self.assertEqual(
            mock.mock_calls,
            [call(sentinel.client, service=LOGGER_SERVICE, default_result=LoggerResult.ERROR)],
        )


class GetCacheTest(TestCase):
    def test_get_cache(self):
        settings = Settings(
            redis_cache_cls="fakeredis.aioredis.FakeRedis",
            redis_cache_options={"some": "options"},
        )
        with patch("fakeredis.aioredis.FakeRedis") as mock:
            mock.return_value = sentinel.cache
            self.assertEqual(get_cache(settings), sentinel.cache)

        self.assertEqual(mock.mock_calls, [call(some="options")])
