from unittest import TestCase
from unittest.mock import sentinel

from fastapi import Request
from fastapi.datastructures import Headers

from inquisitor.utils import get_source_ip, import_path, load_instance

TEST_ATTRIBUTE = sentinel.test_attribute


class TestClass:
    """Test class for load_instance."""

    def __init__(self, *, keyword=None, another=None):
        self.keyword = keyword
        self.another = another


class ImportPathTest(TestCase):
    def test_class(self):
        cls = import_path("inquisitor.tests.test_utils.ImportPathTest")
        self.assertEqual(cls, ImportPathTest)

    def test_attribute(self):
        attribute = import_path("inquisitor.tests.test_utils.TEST_ATTRIBUTE")
        self.assertEqual(attribute, sentinel.test_attribute)

    def test_invalid(self):
        data = (
            # path, error_message
            ("", "is not a valid dotted path"),
            (".", "Empty module name"),
            (".invalid", "Empty module name"),
            ("invalid", "is not a valid dotted path"),
            ("invalid.", "No module named"),
            ("invalid.path", "No module named"),
            ("invalid..path", "No module named"),
            ("inquisitor.tests.invalid.TestCase", "No module named"),
            ("inquisitor.tests.test_utils.", "does not have a class or attribute"),
            ("inquisitor.tests.test_utils.InvalidTestCase", "does not have a class or attribute"),
        )
        for path, error in data:
            with self.subTest(path=path):
                with self.assertRaisesRegex(ImportError, error):
                    import_path(path)


class LoadInstanceTest(TestCase):
    def test_class_only(self):
        obj = load_instance("inquisitor.tests.test_utils.TestClass")

        self.assertIsInstance(obj, TestClass)
        self.assertIsNone(obj.keyword)
        self.assertIsNone(obj.another)

    def test_class_kwargs(self):
        obj = load_instance(
            "inquisitor.tests.test_utils.TestClass",
            keyword=sentinel.value,
            another=sentinel.another,
        )

        self.assertIsInstance(obj, TestClass)
        self.assertEqual(obj.keyword, sentinel.value)
        self.assertEqual(obj.another, sentinel.another)

    def test_invalid_path(self):
        data = (
            # path, error_message
            ("", "is not a valid dotted path"),
            ("inquisitor.tests.test_utils.InvalidTestCase", "does not have a class or attribute"),
        )
        for path, error in data:
            with self.subTest(path=path):
                with self.assertRaisesRegex(ImportError, error):
                    load_instance(path)

    def test_invalid_kwargs(self):
        data = (
            {"invalid": sentinel.invalid},
            {
                "keyword": sentinel.keyword,
                "another": sentinel.another,
                "invalid": sentinel.invalid,
            },
        )
        for kwargs in data:
            with self.subTest(kwargs=kwargs):
                with self.assertRaises(TypeError):
                    load_instance("inquisitor.tests.test_utils.TestClass", **kwargs)


class GetSourceIpTest(TestCase):
    def test_client_host(self):
        request = Request(
            scope={"type": "http", "headers": Headers({}).raw, "client": ("192.168.0.2", 80)}
        )
        self.assertEqual(get_source_ip(request), "192.168.0.2")

    def test_fallback(self):
        request = Request(scope={"type": "http", "headers": Headers({}).raw})
        self.assertEqual(get_source_ip(request), None)
