import os
from contextlib import contextmanager
from datetime import datetime, timezone
from typing import Any, cast
from unittest import IsolatedAsyncioTestCase as TestCase
from unittest.mock import MagicMock, call, patch, sentinel

from fastapi.testclient import TestClient
from fred_types import ContactHistoryId, ContactId
from grill import LoggerClient, Properties, References
from regal import Contact, ContactClient, DomainClient, Registrar, RegistrarClient
from verifier import (
    ContactData,
    Event,
    Verification,
    VerificationClient,
    VerificationId,
    VerificationResult,
    VerificationStatus,
)

from inquisitor.auth import TokenData, create_access_token
from inquisitor.constants import LOGGER_SERVICE, OAUTH_COOKIE, LogEntryType, LoggerResult, Scope
from inquisitor.main import app
from inquisitor.session import Session, get_session
from inquisitor.settings import (
    get_contact_client,
    get_domain_client,
    get_logger_client,
    get_registrar_client,
    get_settings,
    get_verification_client,
)

VERIFICATION_ID = VerificationId("v:1")
CONTACT_ID = ContactId("c:1")


@contextmanager
def set_environ(**kwargs):
    """Clean the environment and set only provided variables."""
    with patch.dict(os.environ) as mock:
        os.environ.clear()
        os.environ.update(kwargs)
        yield mock


def get_auth_cookies():
    access_token = create_access_token(
        TokenData(verification_id=VERIFICATION_ID, contact_id=CONTACT_ID, scopes=list(Scope)),
        get_settings(),
    )
    return {OAUTH_COOKIE: f"Bearer {access_token}"}


def make_verification(**kwargs: Any) -> Verification:
    """Get verification."""
    defaults = {
        "verification_id": VERIFICATION_ID,
        "contact_id": CONTACT_ID,
        "create_datetime": datetime(2022, 1, 1, tzinfo=timezone.utc),
        "close_datetime": None,
        "status": VerificationStatus.CREATED,
        "events": [],
        "next_events": [],
    }
    defaults.update(**kwargs)

    return Verification(**defaults)


def make_event(**kwargs: Any) -> Event:
    """Get verification event."""
    defaults = {
        "event_datetime": datetime(2023, 1, 1, tzinfo=timezone.utc),
        "type": "create",
        "method": None,
        "details": {},
    }
    defaults.update(**kwargs)

    return Event(**defaults)


def make_verification_result(**kwargs: Any) -> VerificationResult:
    """Get verification result."""
    defaults = {
        "verify_datetime": datetime(2022, 1, 1, tzinfo=timezone.utc),
        "method": "letter",
        "data": ContactData(name="RIMMER"),
        "contact_history_id": ContactHistoryId("ch:1"),
        "name": False,
    }
    defaults.update(**kwargs)

    return VerificationResult(**defaults)


def make_contact(**kwargs: Any) -> Contact:
    """Get contact."""
    defaults = {
        "contact_id": CONTACT_ID,
        "contact_history_id": ContactHistoryId("ch:1"),
        "contact_handle": "RIMMER",
        "sponsoring_registrar": "REG-JMC",
        "name": "Arnold Rimmer",
        "organization": "Jupiter Mining Corporation",
    }
    defaults.update(**kwargs)

    return Contact(**defaults)


def make_registrar(**kwargs: Any) -> Registrar:
    """Get registrar."""
    defaults = {
        "registrar_id": "REG-JMC",
        "registrar_handle": "REG-JMC",
        "name": "Jupiter Mining Corporation",
        "organization": "Jupiter Mining Corporation l.t.d",
        "url": "https://jmc.example.com",
    }
    defaults.update(**kwargs)

    return Registrar(**defaults)


class AuthMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.client = TestClient(app, cookies=get_auth_cookies(), raise_server_exceptions=True)


class ContactClientMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.contact_client = MagicMock(spec=ContactClient)
        app.dependency_overrides[get_contact_client] = lambda: self.contact_client

    def tearDown(self):
        cast(TestCase, super()).tearDown()
        app.dependency_overrides = {}


class DomainClientMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.domain_client = MagicMock(spec=DomainClient)
        app.dependency_overrides[get_domain_client] = lambda: self.domain_client

    def tearDown(self):
        cast(TestCase, super()).tearDown()
        app.dependency_overrides = {}


class RegistrarClientMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.registrar_client = MagicMock(spec=RegistrarClient)
        app.dependency_overrides[get_registrar_client] = lambda: self.registrar_client

    def tearDown(self):
        cast(TestCase, super()).tearDown()
        app.dependency_overrides = {}


class VerificationClientMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.verification_client = MagicMock(spec=VerificationClient)
        app.dependency_overrides[get_verification_client] = lambda: self.verification_client

    def tearDown(self):
        cast(TestCase, super()).tearDown()
        app.dependency_overrides = {}


class LoggerMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.logger_client = MagicMock(spec=LoggerClient)
        self.logger_client.create_log_entry.return_value = sentinel.log_entry_id
        app.dependency_overrides[get_logger_client] = lambda: self.logger_client

    def tearDown(self):
        cast(TestCase, super()).tearDown()
        app.dependency_overrides = {}

    def assertLog(
        self,
        log_entry_type: LogEntryType,
        result: LoggerResult,
        *,
        input_properties: Properties | None = None,
        output_properties: Properties | None = None,
        input_references: References | None = None,
        output_references: References | None = None,
        input_content: str | None = None,
        output_content: str | None = None,
    ) -> None:
        cast(TestCase, self).assertEqual(
            self.logger_client.mock_calls,
            [
                call.__hash__(),  # because of @lru_cache on get_logger dependency
                call.create_log_entry(
                    LOGGER_SERVICE,
                    log_entry_type,
                    source_ip="testclient",
                    session_id=None,
                    content=input_content,
                    properties=input_properties,
                    references=input_references,
                ),
                call.close_log_entry(
                    sentinel.log_entry_id,
                    result,
                    session_id=None,
                    content=output_content,
                    properties=output_properties or {},
                    references=output_references or {},
                ),
            ],
        )


class SessionMixin:
    def setUp(self):
        cast(TestCase, super()).setUp()
        self.session = Session(verification_id=VERIFICATION_ID)
        app.dependency_overrides[get_session] = lambda: self.session

    def tearDown(self):
        cast(TestCase, super()).tearDown()
        app.dependency_overrides = {}


class ApiTestCase(
    AuthMixin,
    LoggerMixin,
    ContactClientMixin,
    DomainClientMixin,
    RegistrarClientMixin,
    VerificationClientMixin,
    SessionMixin,
    TestCase,
):
    """FastAPI test case."""
